#
# $Copyright: Copyright (c) 2015 Symantec Corporation.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# SYMANTEC CORPORATION.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SYMANTEC CORPORATION.
#
# The Licensed Software and Documentation are deemed to be commercial
# computer software as defined in FAR 12.212 and subject to restricted
# rights as defined in FAR Section 52.227-19 "Commercial Computer
# Software - Restricted Rights" and DFARS 227.7202, "Rights in
# Commercial Computer Software or Commercial Computer Software
# Documentation", as applicable, and any successor regulations. Any use,
# modification, reproduction release, performance, display or disclosure
# of the Licensed Software and Documentation by the U.S. Government
# shall be solely in accordance with the terms of this Agreement.  $
# $Revision: 1.1.2.34 $
# $Source: /project/arrayrep/src/recoverpoint/unix/Attic/RecoverPointAgent.pm,v $
# $Id: RecoverPointAgent.pm,v 1.1.2.34 2015/05/28 14:08:03 dsathe Exp $

#max message id 66
package RecoverPointAgent;
use ArrayRep;
use ag_i18n_inc;
@ISA = ("ArrayRep");

my $RecoverPointCmd;

use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %default);

my $VERIFY_GROUP_TIMEOUT = 10;

%default = {
	start_time=>'',
	resname => '',
	rpaaddr => '',
	rpauser => '',
	ssh_bin => '',
	ssh_path_to_id_file => '',
	cgname => '',
	copyname => '',
	failover_image => (),
	start_transfer => '',
	trymigrate => '',
	agentopts => (),
	vcsbookmark => '',
	vcsretrymigrate => '',
	production_copy => '',
	version=>''
};

sub new {
	my $proto = shift;
	my $resname = shift;
	my @attrs = @_;
	my $class = ref($proto) || $proto;
	my $self = { %default };
	$self->{'resname'} = $resname;
	bless($self, $class);
	my $ret = $VCSAG_SUCCESS;
	$self->{'start_time'} = $^T;
	
	
	($ret, $self->{'rpaaddr'}) = VCSAG_GET_ATTR_VALUE ("RPAAddr", -1, 1, @attrs);
	exit 0 if ($ret != $VCSAG_SUCCESS);

	($ret, $self->{'rpauser'}) = VCSAG_GET_ATTR_VALUE ("RPAUser", -1, 1, @attrs);
	exit 0 if ($ret != $VCSAG_SUCCESS);

	($ret, $self->{'ssh_bin'}) = VCSAG_GET_ATTR_VALUE ("SSHBinary", -1, 1, @attrs);
exit 0 if ($ret != $VCSAG_SUCCESS);

	($ret, $self->{'ssh_path_to_id_file'}) = VCSAG_GET_ATTR_VALUE ("SSHPathToIDFile", -1, 1, @attrs);
	exit 0 if ($ret != $VCSAG_SUCCESS);

	($ret, $self->{'cgname'}) = VCSAG_GET_ATTR_VALUE ("ConsistencyGroupName", -1, 1, @attrs);
	exit 0 if ($ret != $VCSAG_SUCCESS);

	($ret, $self->{'copyname'}) = VCSAG_GET_ATTR_VALUE ("CopyName", -1, 1, @attrs);
	exit 0 if ($ret != $VCSAG_SUCCESS);
	
	$ret = $VCSAG_SUCCESS;
	my $indx = 1;
	my $val;
	$self->{'failover_image'} = ();
	($ret, $val) = VCSAG_GET_ATTR_VALUE ("FailoverImage", -1, $indx, @attrs);
	while ($ret == $VCSAG_SUCCESS) {
		push (@{$self->{'failover_image'}}, $val);
		$indx++;
		($ret, $val) = VCSAG_GET_ATTR_VALUE ("FailoverImage", -1, $indx, @attrs);
	}
	
	($ret, $self->{'trymigrate'}) = VCSAG_GET_ATTR_VALUE ("TryMigrate", -1, 1, @ARGV);
	exit 0 if ($ret != $VCSAG_SUCCESS);
	
	($ret, $self->{'start_transfer'}) = VCSAG_GET_ATTR_VALUE ("StartTransfer", -1, 1, @ARGV);
	exit 0 if ($ret != $VCSAG_SUCCESS);


	$ret = $VCSAG_SUCCESS;
	my $indx = 1;
	my $val;
	$self->{'agentopts'} = ();
	($ret, $val) = VCSAG_GET_ATTR_VALUE ("AgentOpts", -1, $indx, @attrs);
	while ($ret == $VCSAG_SUCCESS) {
		push (@{$self->{'agentopts'}}, $val);
		$indx++;
		($ret, $val) = VCSAG_GET_ATTR_VALUE ("AgentOpts", -1, $indx, @attrs);
	}

	($ret, $self->{'vcsbookmark'}) = VCSAG_GET_ATTR_VALUE ("VCSBookMark", -1, 1, @attrs);
	exit 0 if ($ret != $VCSAG_SUCCESS);
	
	($ret, $self->{'vcsretrymigrate'}) = VCSAG_GET_ATTR_VALUE ("VCSRetryMigrate", -1, 1, @attrs);
	exit 0 if ($ret != $VCSAG_SUCCESS);

	return $self;
}

#validates agent arguments
#RC = 0, validation succeeded
#RC = 1, validation failed
sub validate_args {
	my $self = shift;
	my $ret = 0;
	
	if ($self->{'rpaaddr'} eq "") {
		VCSAG_LOG_MSG("E", "RPAAddr (name/IP) must be specified.", 1);
		$ret = 1;
	}
	if ($self->{'rpauser'} eq "") {
		VCSAG_LOG_MSG("E", "RPAUser must be specified.", 2);
		$ret = 1;
	}
	if ($self->{'ssh_bin'} eq "") {
		VCSAG_LOG_MSG("E", "SSHBinary must be specified.", 3);
		$ret = 1;
	}
	if ($self->{'ssh_path_to_id_file'} eq "") {
		VCSAG_LOG_MSG("E", "SSHPathToIDFile must be specified.", 4);
		$ret = 1;
	}
	if ($self->{'cgname'} eq "") {
		VCSAG_LOG_MSG("E", "ConsistencyGroupName must be specified.", 5);
		$ret = 1;
	}
	if ($self->{'copyname'} eq "") {
		VCSAG_LOG_MSG("E", "CopyName must be specified.", 6);
		$ret = 1;
	}
	if (($self->{'trymigrate'} != 0) && ($self->{'trymigrate'} != 1)) {
		VCSAG_LOG_MSG("E", "TryMigrate must have value of 0 or 1.", 7);
		$ret = 1;
	}
	if (($self->{'start_transfer'} != 0) && ($self->{'start_transfer'} != 1)) {
		VCSAG_LOG_MSG("E", "StartTransfer must have value of 0 or 1.", 8);
		$ret = 1;
	}
		
	if ($#{$self->{'failover_image'}} != 1) {
		VCSAG_LOG_MSG("E", "FailoverImage format is incorrect.", 9);
		$ret = 1;
	} elsif (($self->{'failover_image'}[0] ne "TIME") && ($self->{'failover_image'}[0] ne "IMAGE")) {
		VCSAG_LOG_MSG("E", "FailoverImage format is incorrect.", 9);
		$ret = 1;
	} elsif ($self->{'failover_image'}[1] eq "") {
		VCSAG_LOG_MSG("E", "FailoverImage format is incorrect, please specify the image.", 10);
		$ret = 1;
	}
	
	return $ret;
}

#Executes the given command on the RPA
#RC >= 0, return code of the RP command
#RC < 0, cannot determine the return code of the RP command
sub rpa_cmd {
	my $self = shift;
	my $cmd = shift;
	my $ret;
        my $rpa_cmd;
	if ($^O =~ m/MSWin32/i ) {
		$rpa_cmd = "\"" . $self->{'ssh_bin'}."\" -l " . $self->{'rpauser'} .
						" -i \"" . $self->{'ssh_path_to_id_file'} . "\" " 
						. $self->{'rpaaddr'} . " " . $cmd . ";get_return_code " . "2>&1";
        }
	else
	{
		$rpa_cmd = $self->{'ssh_bin'}." -l " . $self->{'rpauser'} .	" -i " . $self->{'ssh_path_to_id_file'} . "  " 
						. $self->{'rpaaddr'} . " " . "\"". $cmd . ";get_return_code" ."\"". " 2>&1";
	}

        my $rpa_cmd_log = $rpa_cmd;
        my $rpa_cmd_log_param = $rpa_cmd;

	if ($^O =~ m/MSWin32/i ) {
		$rpa_cmd_log =~ s/"/\\"/g;
		$rpa_cmd_log =~ s/\\"//g;
        }
        $rpa_cmd_log_param =~ s/"/\\"/g; 
         
        VCSAG_LOG_MSG("I", "Executing command [$rpa_cmd_log]", 11, "$rpa_cmd_log_param"); 
	my $out = `$rpa_cmd`;
	my @out = split('\n', $out);
	@out = reverse(@out);
	foreach my $line (@out) {
		if ($line =~ /\S/) {
			if ($line =~ /[^0-9]/) {
				$ret = -1;
			} else {
				$ret = $line;
			}
			last;
		}
	}
	if ($ret == -1) {
		VCSAG_LOG_MSG("E", "Failed to get return code of RP command. Returning -1.", 12); 
	}
	my $out_log = $out;
        if ($^O =~ m/MSWin32/i ) {
	    $out_log =~ s/"/\\"/g;
	    $out_log =~ s/\n/\\n/g;
        } 
	if ($ret != 0) {
		VCSAG_LOG_MSG("E", "Command Failed output [$out_log]", 65, "$out_log"); 
		VCSAG_LOG_MSG("E", "Command exit code [$ret]", 66, "$ret"); 
	} else {
		VCSAG_LOGDBG_MSG(1, "Command output [$out_log]", 12,  "$out_log");
		VCSAG_LOGDBG_MSG(1, "Command exit code [$ret]", 13, "$ret");
	}
	return ($ret, $out);
	
}

sub rpcmd_unexp_error() {
	VCSAG_LOG_MSG("E", "RP command failed." .
						"Administrative intervention is required.", 14); 
}

sub DESTROY {}

####
#RecoverPoint functionality related subs
####

#Get RecoverPoint version
sub get_version {
	my $self = shift;
	my $version;
	my ($ret, $out) = $self->rpa_cmd("get_version");

	if ($ret != 0) {
		VCSAG_LOG_MSG("E", "RP command failed. Cannot determine RPA version.", 63);
		return undef;
	}
	
	my $line = $out;
	if ($line =~ /Version:\s+(\d+.\d+).*/) {
	#Version: 4.1.SP1.P1(h.167)
		VCSAG_LOGDBG_MSG(1, "RecoverPoint version $1 is detected.", $1);
		return $1;		
	} else {
		VCSAG_LOG_MSG("W", "Failed to find RecoverPoint version. Considering RecoverPoint version as 4.1. Please note that agent may fail for prior RecoverPoint version 4.1", 64);
		$version = "4.1";
		return  $version;
	}
	return undef;
}

#tests whether the current site copy is a replica copy
#RC = 0, the current site copy is a production copy
#RC = 1, the current site copy is a replica copy
#RC = 2, cannot determine current site copy status
sub is_replica {
	my $self = shift;

	my ($ret, $out) = $self->rpa_cmd("get_group_settings group='$self->{'cgname'}'"); 
					
	if ($ret != 0) {
		VCSAG_LOG_MSG("E", "RP command failed. Cannot determine role of current site copy.", 15);
		return 2;
	}
	
	my @out = split('\n', $out);
	my $found = 0;
	my $role = "";
	my $prod_copy_found = 0;
	foreach my $line (@out) {
	
		if ((!$prod_copy_found) && ($line =~ /^\s*Production Copy:/)) {
			my(@prod_copy_line) = split(': ', $line);
			$self->{'production_copy'} = $prod_copy_line[1];
			$prod_copy_found = 1;
			VCSAG_LOG_MSG("I", "Production Copy is [$self->{'production_copy'}]", 62, "$self->{'production_copy'}");
			next;
		}
		
		if ($line =~ /^\s*\Q$self->{'copyname'}\E:\s*$/) {
			$found = 1;
			next;
		}
		if ($found) {
				if ($line =~ /^\s*Role:\s*(\S.*)/) {
					$role = $1;
					last;
				}
		}
	}
	VCSAG_LOG_MSG("I", "Role is [$role]", 16, "$role");
	
	if (($role eq "PRODUCTION") or ($role eq "SOURCE (non-production)")) {
		VCSAG_LOG_MSG("I", "Current site copy is the production copy.", 17);
		return 0;
	}
	
	if (($role eq "REMOTE REPLICA") or ($role eq "REMOTE COPY") or ($role eq "LOCAL COPY")) {
		VCSAG_LOG_MSG("I", "Current site copy is the replica copy.", 18);
		return 1;
	}
	
	VCSAG_LOG_MSG("E", "Cannot determine role of current site copy.", 19);
	return 2;
}

#gets the role of the current site copy
sub get_role {
	my $self = shift;
	my $role = "unknown";
	my ($ret, $out) = $self->rpa_cmd("get_group_settings group='$self->{'cgname'}'"); 
					
	if ($ret != 0) {
		VCSAG_LOG_MSG("E", "RP command failed. Cannot determine role of current site copy.", 15);
		return $role;
	}
	
	my @out = split('\n', $out);
	my $found = 0;
	foreach my $line (@out) {
		if ($line =~ /^\s*\Q$self->{'copyname'}\E:\s*$/) {
			$found = 1;
		}
		if ($found) {
				if ($line =~ /^\s*Role:\s*(\S.*)/) {
					$role = $1;
					last;
				}
		}
	}
	VCSAG_LOG_MSG("I", "Role is [$role]", 16, "$role");
	return $role;
}

#gets state of data transfer
sub get_transfer_state {
	my $self = shift;
	my ($ret, $out) = $self->rpa_cmd("get_group_state group='$self->{'cgname'}'"); 
					
	if ($ret != 0) {
		VCSAG_LOG_MSG("E", "RP command failed. Cannot determine state of data transfer.", 20);
		return "unknown";
	}
	
	my @out = split('\n', $out);
	my $found = 0;
	my $state = "unknown";
	foreach my $line (@out) {
		if ($line =~ /^\s*Data Transfer:\s*(\S.*)/) {
			$state = $1;
			last;
		}
	}
	return $state;
}

#tests whether the data transfer is paused
#RC = 0, the data transfer is paused
#RC = 1, the data transfer is not paused
#RC = 2, cannot determine the state of data transfer
sub is_paused {
	my $self = shift;
	my ($ret, $out) = $self->rpa_cmd("get_group_state group='$self->{'cgname'}'"); 
					
	if ($ret != 0) {
		VCSAG_LOG_MSG("E", "RP command failed. Cannot determine state of data transfer.", 21);
		return 2;
	}
	
	my @out = split('\n', $out);
	my $found = 0;
	my $state = "";
	foreach my $line (@out) {
		if ($line =~ /^\s*Data Transfer:\s*(\S.*)/) {
			$state = $1;
			last;
		}
	}
	VCSAG_LOG_MSG("I", "State of data transfer is [$state]", 22, "$state");
	
	if ($state eq "PAUSED") {
		VCSAG_LOG_MSG("I", "Data transfer is currently paused.", 23);
		return 0;
	} elsif ($state ne "") {
		VCSAG_LOG_MSG("I", "Data transfer is not currently paused.", 24);
		return 1;
	}
	
	VCSAG_LOG_MSG("E", "Cannot determine state of data transfer.", 25);
	return 2;
}

#tests whether an image is already being accessed
#RC = 0, no image is being accessed
#RC = 1, image is being accessed
#RC = 2, cannot determine
sub is_image_accessed() {
	my $self = shift;
	my ($ret, $out) = $self->rpa_cmd("verify_group group='$self->{'cgname'}'" . 
					" copy='$self->{'copyname'}' target_image=distributing timeout=$VERIFY_GROUP_TIMEOUT"."secs");
	if (($ret != 0)  && ($ret != 1)) {
		#something unexpected
		$self->rpcmd_unexp_error();
		VCSAG_LOG_MSG("E", "Cannot determine whether an image is being accessed.", 26); 
		return 2;
	}
	if ($ret == 0) {
		VCSAG_LOG_MSG("I", "No image is being accessed.", 27);
	} else {
		VCSAG_LOG_MSG("I", "Image is being accessed.", 28);
	}
	return $ret;
}

#starts the data transfer
#RC = 0, starting data transfer succeeded
#RC = 1, failed
sub start_transfer() {
	my $self = shift;
	my ($ret, $out) = $self->rpa_cmd("start_transfer group='$self->{'cgname'}'"); 
	if ($ret != 0) {
		return 1;
	} else {
		return $ret;
	}
}

#pauses the data transfer
#RC = 0, pausing data transfer succeeded
#RC = 1, failed
sub pause_transfer() {
	my $self = shift;
	my ($ret, $out) = $self->rpa_cmd("pause_transfer group='$self->{'cgname'}'"); 
	if ($ret != 0) {
		return 1;
	} else {
		return $ret;
	}
}

#Checks whether specified bookmark exists
#RC = 0, bookmark exists
#RC = 1, bookmark does not exist
#RC = 2, cannot determine
sub bookmark_exists {
	my $self = shift;
	my $name = shift;
	my ($ret, $out) = $self->rpa_cmd("verify_group group='$self->{'cgname'}'" . 
					" copy='$self->{'copyname'}' bookmark_arrived='$name' timeout=$VERIFY_GROUP_TIMEOUT"."secs");
	if (($ret != 0)  && ($ret != 1)) {
		#exit since something unexpected
		$self->rpcmd_unexp_error(); 
		return 2;
	}
	if ($ret) {
		VCSAG_LOG_MSG("I", "Bookmark [$name] doesn't exist currently.", 29, "$name"); 
	} else {
		VCSAG_LOG_MSG("I", "Bookmark [$name] exists currently.", 30, "$name"); 
	}
	return $ret;
}

#Enables access to image
#RC = 0, succeeded
#RC = 1, failed
#RC = 2, cannot determine
sub enable_image_access {
	my $self = shift;
	my $image = shift;
	my $image_cli_opt;
	my $latest_image;
	
	if ($self->{'version'} == "") {
		$self->{'version'} = $self->get_version();
	}
	if ($self->{'version'} < 4.1) {
		$latest_image = 'latest';
	} else {
		$latest_image = 'latest_image';
	}
	if (!defined $image)  {
		if ($self->{'failover_image'}[0] eq "IMAGE" ) {
			#IMAGE=
			if (($self->{'failover_image'}[1] eq "latest") ||
				($self->{'failover_image'}[1] eq "latest_image"))
			{
				$image_cli_opt = "image=$latest_image";
			} else {
				#validate if bookmark is valid
				if ($self->bookmark_exists($self->{'failover_image'}[1]) > 0) {
					VCSAG_LOG_MSG("E", "Cannot validate that bookmark exists currently.", 31);
					return 1;
				} else {
					if ($self->{'version'} < 4.1) {
                        			$image_cli_opt = "image='" . $self->{'failover_image'}[1] . "'";
                			} else {
                        			$image_cli_opt = "image='search_image' search_text='" . $self->{'failover_image'}[1] . "'";
                			}
				}
			}
		} else {
			#TIME=
			$image_cli_opt = "time='" . $self->{'failover_image'}[1] . "'";
		}
			
	} else {
		if ($self->{'version'} < 4.1) {
                        $image_cli_opt = "image='" . $image . "'";
                } else {
                        $image_cli_opt = "image='search_image' search_text='" . $image . "'";
                }
	}
	my ($ret, $out) = $self->rpa_cmd("enable_image_access group='$self->{'cgname'}'" . 
					" copy='$self->{'copyname'}' $image_cli_opt ");
	if ($ret > 0) {
			VCSAG_LOG_MSG("E", "Enabling Logged Access to Image failed.", 32);
			return 1;
	}
	if ($ret < 0) {
		return 2;
	}
	return $ret;
}

#disables access to image
#RC = 0, succeeded
#RC = 1, failed
#RC = 2, cannot determine
sub disable_image_access {
	my $self = shift;
	my ($ret, $out) = $self->rpa_cmd("disable_image_access group='$self->{'cgname'}'" . 
					" copy='$self->{'copyname'}' ");
	if ($ret > 0) {
			VCSAG_LOG_MSG("E", "Disabling Logged Access to image failed.", 33);
			return 1;
	}
	if ($ret < 0) {
		return 2;
	}

	return $ret;
}

#gets group state
#RC = 0, succeeded
#RC = 1, failed
#RC = 2, cannot determine
sub get_group_state {
	my $self = shift;
	my ($ret, $out) = $self->rpa_cmd("get_group_state group='$self->{'cgname'}'");
	if ($ret > 0) {
			VCSAG_LOG_MSG("E", "Getting group state failed.", 34);
			return 1;
	}
	if ($ret < 0) {
		return 2;
	}

	return ($ret, $out);
}

#wait until image access is complete
#RC = 0, image access completed
#RC = 1, timed out
#RC = 2, other error
sub wait_for_image_access {
	my $self = shift;
	my $oto = $self->get_oto();
	my $curr_time = time();
	my $timeout = $oto - ($curr_time - $self->{'start_time'}) - 20;
	if ($timeout <= 0) {
		VCSAG_LOG_MSG("E", "Computed value of timeout  [$timeout] is <= 0." .
		"Please increase the value of OnlineTimeout.", 35, "$timeout");
		return 2;
	}
	my ($ret, $out) = $self->rpa_cmd("verify_group group='$self->{'cgname'}'" . 
					" copy='$self->{'copyname'}' target_image=logged_access timeout=$timeout"."secs");
	if (($ret != 0)  && ($ret != 1)) {
		#exit since something unexpected
		$self->rpcmd_unexp_error(); 
		return 2;
	}
	if ($ret) {
		VCSAG_LOG_MSG("E", "Timed out waiting for image access to complete.", 36); 
	} else {
		VCSAG_LOG_MSG("I", "Image access has completed.", 37); 
	}
	return $ret;
}

#performs permanent failover to CRR replica copy
#RC = 0, failover succeeded
#RC = 1, failover did not succeed
#RC = 2, cannot determine
sub failover {
	my $self = shift;
	my $opts ;
	my ($ret, $out);
	
	#get number of copies configure for CG
	my $copy_count = $self->get_copies_count();
	if (!($copy_count ==  2 || $copy_count == 3)) {
		VCSAG_LOG_MSG("E", "Group $self->{'cgname'} has [$copy_count] RecoverPoint copies. Supported copies are either 2 or 3",61, "$self->{'cgname'}", "$copy_count"); 
		return 1;
	}

	if ($self->{'start_transfer'} == 0) {
		$opts = "start_transfer=no";
	} else {
		$opts = "start_transfer=yes";
	}
	
	if (($copy_count == 3) && ($self->{'production_copy'} eq $self->{'copyname'})) {
		($ret, $out) = $self->rpa_cmd("resume_production group='$self->{'cgname'}' ". $opts);
	} else {
		($ret, $out) = $self->rpa_cmd("failover group='$self->{'cgname'}'" . 
					" copy='$self->{'copyname'}' -f " . $opts);
	}

	if ($ret > 0) {
		VCSAG_LOG_MSG("E", "Failing over to the replica copy failed.", 38);
		return 1;
	} elsif ($ret == 0) {
		VCSAG_LOG_MSG("I", "Failing over to the replica copy succeeded.", 39);
		sleep(5);
		return 0;
	}
	return 2;
}

#applies the bookmark
#RC = 0, success
#RC = 1, failed
#RC = 2,cannot determine
sub bookmark_image() {
	my $self = shift;
	my $vcs_bmk_name = shift;
	my ($ret, $out);
	
	if ($self->{'version'} == "") {
		$self->{'version'} = $self->get_version();
	}

	if ($self->{'version'} < 4.1) {
		($ret, $out) = $self->rpa_cmd("bookmark_image group='$self->{'cgname'}' bookmark='$vcs_bmk_name'");
	} else {
		($ret, $out) = $self->rpa_cmd("bookmark_image groups='$self->{'cgname'}' bookmark='$vcs_bmk_name'");
	}

	if ($ret > 0) {
		VCSAG_LOG_MSG("E", "Applying bookmark failed.", 40);
		return 1;
	} elsif ($ret == 0) {
		VCSAG_LOG_MSG("I", "Applying bookmark succeeded.", 41);
		return 0;
	} 
	return 2;
}

#wait until bookmark arrives
#RC = 0, bookmark has arrived
#RC = 1, timed out
#RC = 2, other error
sub wait_for_bookmark {
	my $self = shift;
	my $vcs_bmk_name = shift;
	my $oto = $self->get_oto();
	my $curr_time = time();
	my $timeout = $oto - ($curr_time - $self->{'start_time'}) - 60;
	if ($timeout <= 0) {
		VCSAG_LOG_MSG("E", "Computed value of timeout  [$timeout] is <= 0." .
		"Please increase the value of OnlineTimeout.", 42, "$timeout");
		return 2;
	}
	my ($ret, $out) = $self->rpa_cmd("verify_group group='$self->{'cgname'}'" . 
					" copy='$self->{'copyname'}' bookmark_arrived='$vcs_bmk_name' timeout=$timeout"."secs");
	if (($ret != 0)  && ($ret != 1)) {
		#exit since something unexpected
		$self->rpcmd_unexp_error(); 
		return 2;
	}
	if ($ret) {
		VCSAG_LOG_MSG("E", "Timed out waiting for bookmark to arrive.", 43); 
	} else {
		VCSAG_LOG_MSG("I", "Bookmark has arrived.", 44); 
	}
	return $ret;
}

sub exit_after_failover() {
	my $self = shift;
	VCSAG_LOG_MSG("E", "Administrative intervention required since the agent couldn't complete failover." .
				"Exiting the entry point. The group has been flushed and frozen.", 60); 
	#flush and freeze
	$self->flush_freeze();
}

sub exit_after_enable() {
	my $self = shift;
	VCSAG_LOG_MSG("E", "Exiting the entry point. The group has been flushed and frozen.".
		"Please manually verify that image access has completed, run the failover action and retry online.", 45); 
	#flush and freeze
	$self->flush_freeze();
}

sub exit_migrate_timeout() {
	my $self = shift;
	VCSAG_LOG_MSG("E", "Exiting the entry point. The group has been flushed and frozen.".
		"Migration attempt has timed out waiting for the applied bookmark to arrive." .
		"Consider increasing the OnlineTimeout to avoid this issue in future." .
		"If you wish to retry the same migration attempt that has timed out, " . 
		"execute the set_retry_migrate_flag action and retry online of the RecoverPoint resource on the same host." .
		"If you have executed the set_retry_migrate_flag action and subsequently do not wish to online the RecoverPoint " . 
		" resource, then make sure you execute the unset_retry_migrate_flag action.", 46); 
	#flush and freeze
	$self->flush_freeze();
}

####
#Business logic
####

#Attempts to failover in RecoverPoint CRR
#RC = 0, failover succeeded
#RC = 1, failover did not succeed
sub crr_failover {
	my $self = shift;
	
	#enable image access
	if ($self->enable_image_access() != 0) {
			return 1;
	}
	
	#wait until image is accessed
	if ($self->wait_for_image_access() != 0) {
		$self->exit_after_enable();
		return 1;
	}

	#failover
	if ($self->failover() != 0) {
		$self->exit_after_failover();
		return 1;
	}
	return 0;
}

#Attempts to do a migration in RecoverPoint CRR
#RC = 0, migrate succeeded
#RC = 1, migrate did not succeed
sub crr_migrate {
	my $self = shift;
	my $curr_time = time();
	my $vcs_bmk_name;
	
	#should we retry prev. migration attempt ?
	if ($self->{'vcsretrymigrate'} == 0) {
		my $sys = get_sys();
		$vcs_bmk_name = "VCS_Migrate_" . $sys . "_" . $curr_time;
		VCSAG_LOG_MSG("I", "Generated VCS Bookmark name [$vcs_bmk_name] .", 47, "$vcs_bmk_name");
		#bookmark the current image
		if ($self->bookmark_image($vcs_bmk_name) != 0) {
				return 1;
		}
		$self->update_bookmark($vcs_bmk_name);	
	} else {
		$vcs_bmk_name = $self->{'vcsbookmark'};
		if ($vcs_bmk_name eq "") {
			VCSAG_LOG_MSG("E", "No previous bookmark found to retry previous migration attempt.", 48); 
			return 1;
		}
		VCSAG_LOG_MSG("I", "Using previous bookmark [$vcs_bmk_name] .", 49, "$vcs_bmk_name");
	}

	#wait for bookmark to arrive
	if ($self->wait_for_bookmark($vcs_bmk_name) != 0) {
		$self->exit_migrate_timeout();
		return 1;
	}
	
	#enable image access
	if ($self->enable_image_access() != 0) {
			return 1;
	}
	
	#wait until image is accessed
	if ($self->wait_for_image_access() != 0) {
		$self->exit_after_enable();
		return 1;
	}

	#failover
	if ($self->failover() != 0) {
		$self->exit_after_failover();
		return 1;
	}
	
	return 0;
}

####
#HA commands related subs
####

sub get_sys {
		my (@fields) = split('\s+', `$ag_i18n_inc::HASYS -nodeid`);
		my $sys = $fields[5];
		return $sys;
	
}

sub clear_vcsretrymigrate {
	my $self = shift;
	my $ret = 0;
	if ($self->{'vcsretrymigrate'} == 1) {
		`$ag_i18n_inc::HARES -modify $self->{'resname'} VCSRetryMigrate 0`;
		$ret = $?;
		if ($ret) {
			VCSAG_LOG_MSG("E", "Failed to clear VCSRetryMigrate.", 50);
		}
	}
	return $ret;

}

sub set_vcsretrymigrate {
	my $self = shift;
	my $ret = 0;
	if ($self->{'vcsretrymigrate'} == 0) {
		`$ag_i18n_inc::HARES -modify $self->{'resname'} VCSRetryMigrate 1`;
		$ret = $?;
		if ($ret) {
			VCSAG_LOG_MSG("E", "Failed to set VCSRetryMigrate.", 51);
		}
	}
	return $ret;
	
}

sub update_bookmark {
my $self = shift;
my $vcs_bmk_name = shift;

	`$ag_i18n_inc::HARES -modify $self->{'resname'} VCSBookMark $vcs_bmk_name`;
	my $ret = $?;
	if ($ret) {
		VCSAG_LOG_MSG("E", "Failed to set VCSBookMark.", 52);
	}

}
	
sub get_oto {
	my $self = shift;
	my ($oto) = split('\n', `$ag_i18n_inc::HARES -value $self->{'resname'} OnlineTimeout 2>nul`);
	my $ret = $?;
	if ($ret) {
		($oto) = split('\n', `$ag_i18n_inc::HATYPE -value RecoverPoint OnlineTimeout 2>nul`);
		my $ret = $?;
		if ($ret) {
			$oto = 300;
		}
	}
	VCSAG_LOG_MSG("I", "Obtained value of OnlineTimeout [$oto].", 53, "$oto"); 
	return $oto;
}

sub flush_freeze() {
	my $self = shift;
	#get group
	my ($grp) = split('\n', `$ag_i18n_inc::HARES -value $self->{'resname'} Group`);
	#get sys
	my (@fields) = split('\s+', `$ag_i18n_inc::HASYS -nodeid`);
	my $sys = $fields[5];
	#flush group operation
	`$ag_i18n_inc::HAGRP -flush $grp -sys $sys 2>&1`;
	#freeze group temporarily on local cluster
	`$ag_i18n_inc::HAGRP -freeze $grp 2>&1`;	
}

sub Rep_Tech_RPO() {
	my $self = shift;
	my ($ret, $out) = $self->rpa_cmd("get_group_statistics group='$self->{'cgname'}'");
        if ($ret != 0) {
		VCSAG_LOG_MSG("E", "RPA command get_group_statistics failed. Cannot compute Recovery point.", 54);
                return 1;        
        }

	my @out = split('\n', $out);
        my $Rep_Tech_RPO = "";
	my $line = "";
	my $lag_context = 0;
        while (@out) {
                $line = shift @out;
                if ($line =~ /Lag:/) {
			$Rep_Tech_RPO = "Replication Lag Info:";
                        $lag_context = 1;
                        next;
                }
                if ($lag_context == 1) {
                        if ($line =~ /Time:(.*)/) {
                                $Rep_Tech_RPO .= "\nReplication Lag =$1";
                        }
                        if ($line =~ /Data:(.*)/) {
                                 $Rep_Tech_RPO .= "\nData Lag =$1";
                        }
                        if ($line =~ /Writes:(.*)/) {
                                $Rep_Tech_RPO .= "\nWrites Lag =$1";
                        }
                }
        }
	if ($Rep_Tech_RPO eq "") {
		VCSAG_LOG_MSG("E", "Failed to compute Recovery point. Please check RecoverPoint configuration.", 55);
		return 1;
	}
	print $Rep_Tech_RPO;
	return 0;
}

sub get_copies_count() {
	my $self = shift;
	my $copy_count = 0;
	my @copies = '';

	my ($ret, $out) = $self->rpa_cmd("get_groups -xml");
        if ($ret != 0) {
		VCSAG_LOG_MSG("E", "RPA command get_groups failed. Cannot compute Recovery point copies.", 57);
                return 1;        
        }
	my @out = split ('\n', $out);
	
	for (my $i = 0; $i< scalar @out; $i++) {
		if ($out[$i] =~ /^\s+<GroupClusterCopiesOutput\s+key="$self->{'cgname'}">/) {
				$i++;
			if($out[$i] =~ /name="copies"/) {
				$i++;
				while ($out[$i] !~ /<\/map>/) {
					if ($out[$i] =~ />(.*)<\//) {
						@copies = split(/,/, $1);
						$copy_count +=scalar @copies;
						VCSAG_LOG_MSG("I", "Group $self->{'cgname'} copies =$copy_count: @copies", 58, "$copy_count", "@copies");
					}
					$i++;
				}
			}
		   }
	}
	return $copy_count;
}

